// Copyright. 2019 - 2022 Neiry. All rights reserved.

#pragma once

#include "Core/ccDefines.h"

#ifdef __cplusplus
#include <cstdint>
#define NOEXCEPT noexcept

#define CLC_CLASS_WN(Wrapped, Name) \
    namespace capsule::client {     \
    class Wrapped;                  \
    }                               \
    typedef class capsule::client::Wrapped* Name

#define CLC_STRUCT_WN(Wrapped, Name) \
    namespace capsule::client {      \
    struct Wrapped;                  \
    }                                \
    typedef struct capsule::client::Wrapped const* Name

extern "C" {
#else
#include <stdint.h>
#define NOEXCEPT

#define CLC_CLASS_WN(Wrapped, Name) \
    struct clCWrapped##d;           \
    typedef struct clCWrapped##d* Name

#define CLC_STRUCT_WN(Wrapped, Name) \
    struct clCWrapped##d;            \
    typedef struct clCWrapped##d const* Name

#endif //__cplusplus

#define CLC_STRUCT(Name) \
    struct Name##d;      \
    typedef struct Name##d* Name
/**
 * \brief Reason for client's disconnect
 */
typedef enum clCDisconnectReason {
    clC_DR_UserRequested,
    clC_DR_Destruction,
    clC_DR_FatalError
} clCDisconnectReason;

/**
 * \brief Connection state to Capsule.
 */
typedef enum clCConnectionState {
    clC_CS_Disconnected,
    clC_CS_Connecting,
    clC_CS_Connected,
    clC_CS_Disconnecting
} clCConnectionState;

/**
 * \brief Connection error
 */
typedef enum clCError {
    clC_Error_OK = 0,
    clC_Error_FailedToConnect,
    clC_Error_FailedToInitConnection,
    clC_Error_FailedToInitialize,
    clC_Error_IndividualNFBNotCalibrated,
    clC_Error_UNKNOWN = 99
} clCError;

/**
 * \brief String implementation
 *
 * Object handled in Capsule.
 * Need call free by "clCString_Free(clCString stringPointer)"
 */
CLC_STRUCT(clCString);

/**
 * \brief Used to connect with Capsule.
 *
 * Handle to client used to connect with Capsule, responsible for control.
 */
CLC_CLASS_WN(ClientPrivate, clCClient);

/**
 * \brief Used to receive channel resistances from Capsule.
 *
 * Contains electrode resistances of the device
 */
CLC_STRUCT_WN(Resistance, clCResistances);

/**
 * \brief List of devices found by device locator.
 *
 * Contains information about the devices
 */
CLC_STRUCT(clCDeviceInfoList);
/**
 * \brief Device info list.
 *
 * Contains device information: name, id (MAC) and type
 */
CLC_STRUCT_WN(DeviceInfo, clCDeviceInfo);
/**
 * \brief Device locator.
 *
 * Searches for available devices of chosen type
 */
CLC_CLASS_WN(DeviceLocatorPrivate, clCDeviceLocator);
/**
 * \brief Neiry Device.
 *
 * Controls connection to the device
 */

CLC_CLASS_WN(DevicePrivate, clCDevice);

CLC_CLASS_WN(ClassificationCardioPrivate, clCCardio);

CLC_CLASS_WN(ClassificationMEMSPrivate, clCMEMS);

/**
 * \brief Neiry Device type.
 */
typedef enum clCDeviceType {
    clC_DT_NeiryBand = 0,
    clC_DT_NeiryBuds = 1,
    clC_DT_NeiryHeadphones = 2,
    clC_DT_NeiryImpulse = 3,
    clC_DT_NeiryAny = 4,
    clC_DT_NeiryCallibri = 5,
    clC_DT_DevOnly_SinWave = 100,
    clC_DT_DevOnly_Noise = 101,
    clC_DT_DevOnly_LSL,
    clC_DT_DevOnly_File
} clCDeviceType;

typedef enum clCDeviceMode {
    clC_DM_Resistance,
    clC_DM_Idle,
    clC_DM_PowerDown,
    clC_DM_Signal,
    clC_DM_StartMEMS,
    clC_DM_StopMEMS,
    clC_DM_StartPPG,
    clC_DM_StopPPG
} clCDeviceMode;

/**
 * \brief Session state.
 */
typedef enum clCSessionState {
    clC_SS_Unknown,
    clC_SS_Starting,
    clC_SS_Active,
    clC_SS_Inactive,
    clC_SS_Pausing,
    clC_SS_Paused,
    clC_SS_Stopping
} clCSessionState;

/**
 * \brief Session error.
 */
typedef enum clCSessionError {
    clC_SE_Unknown,
    clC_SE_FailedToStart,
    clC_SE_FailedToStop,
    clC_SE_LostConnection
} clCSessionError;

/**
 * \brief Device connection state.
 */
typedef enum clCDeviceConnectionState {
    clC_SE_Disconnected = 0,
    clC_SE_Connected = 1,
    clC_SE_UnsupportedConnection = 2,
} clCDeviceConnectionState;

/**
 * \brief Client session.
 *
 * Required for classifiers and calibrators
 */
CLC_CLASS_WN(SessionPrivate, clCSession);

/**
 * \brief Client callback without parameters.
 * (Note: ALL callbacks receive their owner as first argument).
 */
CLC_STRUCT(clCClientDelegate);
typedef void (*clCClientHandler)(clCClient);
CL_DLL void clCClientDelegate_Set(clCClientDelegate delegate,
                                  clCClientHandler callback) NOEXCEPT;
/**
 * \brief Client disconnect callback. Not supported in the current version
 */
CLC_STRUCT(clCClientDelegateDisconnectReason);
typedef void (*clCClientHandlerDisconnectReason)(clCClient,
                                                 clCDisconnectReason);
CL_DLL void clCClientDelegateDisconnectReason_Set(
    clCClientDelegateDisconnectReason delegate,
    clCClientHandlerDisconnectReason callback) NOEXCEPT;
/**
 * \brief Client error callback. Not supported in the current version
 */
CLC_STRUCT(clCClientDelegateError);
typedef void (*clCClientHandlerError)(clCClient, clCError);
CL_DLL void clCClientDelegateError_Set(clCClientDelegateError delegate,
                                       clCClientHandlerError callback) NOEXCEPT;
/**
 * \brief Device callback for the list of device info.
 */
CLC_STRUCT(clCDeviceLocatorDelegateDeviceInfoList);
typedef void (*clCDeviceLocatorHandlerDeviceInfoList)(clCDeviceLocator,
                                                      clCDeviceInfoList);
CL_DLL void clCDeviceLocatorDelegateDeviceInfoList_Set(
    clCDeviceLocatorDelegateDeviceInfoList delegate,
    clCDeviceLocatorHandlerDeviceInfoList callback) NOEXCEPT;
/**
 * \brief Session error callback. Not supported in the current version
 */
CLC_STRUCT(clCSessionDelegateSessionError);
typedef void (*clCSessionHandlerSessionError)(clCSession, clCSessionError);
CL_DLL void
clCSessionDelegateSessionError_Set(clCSessionDelegateSessionError delegate,
                                   clCSessionHandlerSessionError callback) NOEXCEPT;
/**
 * Get Capsule client version
 *
 * \returns client version string
 */
CL_DLL clCString clCClient_GetVersionString() NOEXCEPT;

/**
 * \brief Free memory of clCString
 */
CL_DLL void clCString_Free(clCString stringPointer) NOEXCEPT;

/**
 * Get ANSI string from clCString
 *
 * \returns const char*
 */
CL_DLL const char* clCString_CStr(clCString stringPointer) NOEXCEPT;

/**
 * Create a Capsule client.
 *
 * \returns client handle
 */
CL_DLL clCClient clCClient_Create() NOEXCEPT;
/**
 * Create a Capsule client with application name.
 * \param clientName application name
 *
 * \returns client handle
 */
CL_DLL clCClient clCClient_CreateWithName(const char* clientName) NOEXCEPT;
/**
 * Create a Capsule client with data directory.
 * \param dataDir path to data directory
 *
 * \returns client handle
 */
CL_DLL clCClient clCClient_CreateWithDataDirectory(const char* dataDir) NOEXCEPT;
/**
 * Create a Capsule client with data directory and application name.
 * \param dataDir path to data directory
 * \param clientName application name
 *
 * \returns client handle
 */
CL_DLL clCClient clCClient_CreateWithDataDirectoryAndName(const char* dataDir, const char* clientName) NOEXCEPT;
/**
 * Destroy Capsule client. Release the handle.
 *
 * \param client client handle
 */
CL_DLL void clCClient_Destroy(clCClient client) NOEXCEPT;
/**
 * Connect to Capsule.
 *
 * \param client client handle
 * \param address Capsule address to connect to. If NULL, the default address
 * "tcp:://127.0.0.1:5666" is in effect.
 * * For embedded Capsule use in-process protocol: "inproc://capsule",
 * * Otherwise use tcp protocol: "tcp:://127.0.0.1:5666".
 * OnConnected event is called when connected.
 */
CL_DLL void clCClient_Connect(clCClient client, const char* address) NOEXCEPT;
/**
 * Connect to Capsule with access token.
 *
 * \param client client handle
 * \param address Capsule address to connect to. If NULL, the default address
 * "tcp:://127.0.0.1:5666" is in effect.
 * * For embedded Capsule use in-process protocol: "inproc://capsule",
 * * Otherwise use tcp protocol: "tcp:://127.0.0.1:5666".
 * OnConnected event is called when connected.
 * \param accessToken access token for communication with platform
 */
CL_DLL void clCClient_ConnectWithAccessToken(clCClient client,
                                             const char* address,
                                             const char* accessToken) NOEXCEPT;
/**
 * Connect to Capsule with access token and platform address.
 *
 * \param client client handle
 * \param address Capsule address to connect to. If NULL, the default address
 * "tcp:://127.0.0.1:5666" is in effect.
 * * For embedded Capsule use in-process protocol: "inproc://capsule",
 * * Otherwise use tcp protocol: "tcp:://127.0.0.1:5666".
 * OnConnected event is called when connected.
 * \param accessToken access token for communication with platform
 * \param platformAddress URL of a platform processing http requests
 */
CL_DLL void clCClient_ConnectWithAccessTokenAndPlatformAddress(clCClient client,
                                                               const char* address, const char* accessToken,
                                                               const char* platformAddress) NOEXCEPT;

/**
 * Get delegate for client RefreshAccessTokenRequested event.
 *
 * \param client client handle
 * \returns client delegate
 */
CL_DLL clCClientDelegate
clCClient_GetOnRefreshAccessTokenRequestedEvent(clCClient client) NOEXCEPT;
/**
 * Refresh access token for data upload
 * \param client client handle
 * \param accessToken access token
 */
CL_DLL void clCClient_RefreshToken(clCClient client, const char* accessToken) NOEXCEPT;
/**
 * Get Capsule time for synchronization with EEG timestamps
 * \return time
 */
CL_DLL uint64_t clCClient_GetTime() NOEXCEPT;
/**
 * Get Capsule time in microseconds
 * \return time in microseconds
 */
CL_DLL uint64_t clCClient_GetTimeMicro() NOEXCEPT;
/**
 * Disconnects from Capsule, stopping any background threads.
 *
 * \param client client handle
 */
CL_DLL void clCClient_Disconnect(clCClient client) NOEXCEPT;
/**
 * Is client connected.
 *
 * \param client client handle
 */
CL_DLL bool clCClient_IsConnected(clCClient client) NOEXCEPT;
/**
 * Is client connecting.
 *
 * \param client client handle
 */
CL_DLL bool clCClient_IsConnecting(clCClient client) NOEXCEPT;
/**
 * Create a Capsule session. Requires a device.
 *
 * \param client client handle
 * \param device device handle
 * \returns session handle
 */
CL_DLL clCSession clCClient_CreateSession(clCClient client, clCDevice device) NOEXCEPT;
/**
 * Get Capsule version string.
 *
 * \param client client handle
 *
 * \returns capsule version as a string ("unknown" if not connected).
 */
CL_DLL clCString clCClient_GetCapsuleVersion(clCClient client) NOEXCEPT;
/**
 * Get Client application name.
 *
 * \param client client handle
 *
 * \returns client name as a string ("unknown" if not connected).
 */
CL_DLL clCString clCClient_GetClientName(clCClient client) NOEXCEPT;
/**
 * Get Capsule data directory.
 *
 * \param client client handle
 *
 * \returns capsule data directory as a string ("" if not applicable).
 */
CL_DLL clCString clCClient_GetDataDirectory(clCClient client) NOEXCEPT;
/**
 * Client is single-threaded.
 * \b Note: Update does all the work and must be called
 * regularly to process events.
 *
 * \param client client handle
 */
CL_DLL void clCClient_Update(clCClient client) NOEXCEPT;
/**
 * Creates locator for Neiry devices of the chosen type.
 *
 * \param client client handle
 * \param deviceType device type
 * \returns device locator handle
 */
CL_DLL clCDeviceLocator clCClient_ChooseDeviceType(clCClient client,
                                                   clCDeviceType deviceType) NOEXCEPT;
/**
 * Get delegate for client connected event.
 *
 * \param client client handle
 * \returns client delegate
 */
CL_DLL clCClientDelegate clCClient_GetOnConnectedEvent(clCClient client) NOEXCEPT;
/**
 * Get delegate for client disconnected event.
 *
 * \param client client handle
 * \returns client disconnect delegate
 */
CL_DLL clCClientDelegateDisconnectReason
clCClient_GetOnDisconnectedEvent(clCClient client) NOEXCEPT;
/**
 * Get delegate for client error event.
 *
 * \param client client handle
 * \returns client error delegate
 */
CL_DLL clCClientDelegateError clCClient_GetOnErrorEvent(clCClient client) NOEXCEPT;

/**
 * Release device locator.
 *
 * \param locator device locator handle
 */
CL_DLL void clCDeviceLocator_Destroy(clCDeviceLocator locator) NOEXCEPT;
/**
 * \brief Find available devices. Search time is in seconds.
 * Non-blocking - OnDevices event is fired when timeout expires.
 *
 * \param locator device locator handle
 * \param searchTime search time in seconds
 */
CL_DLL void clCDeviceLocator_RequestDevices(clCDeviceLocator locator,
                                            int32_t searchTime) NOEXCEPT;
/**
 * Create device for interaction.
 *
 * \param locator device locator handle
 * \param deviceID device ID string
 * \returns device handle
 */
CL_DLL clCDevice clCDeviceLocator_CreateDevice(clCDeviceLocator locator, const char* deviceID) NOEXCEPT;
/**
 * Get delegate for available devices event.
 *
 * \param locator device locator handle
 * \returns device list delegate
 */
CL_DLL clCDeviceLocatorDelegateDeviceInfoList clCDeviceLocator_GetOnDevicesEvent(clCDeviceLocator locator) NOEXCEPT;

/**
 * Get total number of resistance channels.
 *
 * \param resistances resistances handle
 * \returns number of channels
 */
CL_DLL int32_t clCResistances_GetCount(clCResistances resistances) NOEXCEPT;
/**
 * Get channel name by index.
 *
 * \param resistances resistances handle
 * \param index channel index
 * \returns channel name
 */
CL_DLL clCString clCResistances_GetChannelName(clCResistances resistances,
                                               int32_t index) NOEXCEPT;
/**
 * Get channel resistance by index.
 *
 * \param resistances resistances handle
 * \param index channel index
 * \returns resistance value
 */
CL_DLL float clCResistances_GetValue(clCResistances resistances, int32_t index) NOEXCEPT;

/**
 * Get the number of devices.
 *
 * \param devices device list handle
 * \returns count of device list
 */
CL_DLL int32_t clCDeviceInfoList_GetCount(clCDeviceInfoList devices) NOEXCEPT;
/**
 * Get device info from the list.
 *
 * \param devices device list handle
 * \param index info index
 * \returns device info handle
 */
CL_DLL clCDeviceInfo clCDeviceInfoList_GetDeviceInfo(clCDeviceInfoList devices,
                                                     int32_t index) NOEXCEPT;
/**
 * Get device ID (Bluetooth MAC address).
 *
 * \param device device info handle
 * \returns device ID string
 */
CL_DLL clCString clCDeviceInfo_GetID(clCDeviceInfo device) NOEXCEPT;
/**
 * Get device name.
 *
 * \param device device info handle
 * \returns device name string
 */
CL_DLL clCString clCDeviceInfo_GetName(clCDeviceInfo device) NOEXCEPT;
/**
 * Get device type.
 *
 * \param device device info handle
 * \returns device type
 */
CL_DLL clCDeviceType clCDeviceInfo_GetType(clCDeviceInfo device) NOEXCEPT;
/**
 * Get compound device description string.
 *
 * \param device device info handle
 * \returns device description string
 */
CL_DLL clCString clCDeviceInfo_GetDescription(clCDeviceInfo device) NOEXCEPT;

/**
 * Device delegate without parameters. Not supported in the current version
 */
CLC_STRUCT(clCDeviceDelegate);
typedef void (*clCDeviceHandler)(clCDevice);
CL_DLL void clCDeviceDelegate_Set(clCDeviceDelegate delegate,
                                  clCDeviceHandler callback) NOEXCEPT;

/**
 * Device delegate for resistances event.
 */
CLC_STRUCT(clCDeviceDelegateResistances);
typedef void (*clCDeviceHandlerResistances)(clCDevice, clCResistances);
CL_DLL void
clCDeviceDelegateResistances_Set(clCDeviceDelegateResistances delegate,
                                 clCDeviceHandlerResistances callback) NOEXCEPT;
/**
 * Device delegate for battery charge.
 */
typedef unsigned char clCUChar;
CLC_STRUCT(clCDeviceDelegateUChar);
typedef void (*clCDeviceHandlerUChar)(clCDevice, clCUChar);
CL_DLL void clCDeviceDelegateBatteryCharge_Set(clCDeviceDelegateUChar delegate,
                                               clCDeviceHandlerUChar callback) NOEXCEPT;

CLC_STRUCT(clCDeviceDelegateMode);
typedef void (*clCDeviceHandlerMode)(clCDevice, clCDeviceMode);
CL_DLL void clCDeviceDelegateMode_Set(clCDeviceDelegateMode delegate,
                                      clCDeviceHandlerMode callback) NOEXCEPT;
/**
 * Get delegate for device connection state event.
 *
 * \param device device handle
 */
CLC_STRUCT(clCDeviceDelegateDeviceConnectionState);
typedef void (*clCDeviceHandlerDeviceConnectionState)(clCDevice, clCDeviceConnectionState);
CL_DLL clCDeviceDelegateDeviceConnectionState clCDevice_GetOnConnectionStateChangedEvent(clCDevice device) NOEXCEPT;

/**
 * Device connection delegate.
 */
CLC_STRUCT(clCDeviceDelegateConnectionState);
typedef void (*clCDeviceHandlerConnectionState)(clCDevice, clCDeviceConnectionState);
CL_DLL void clCDeviceDelegateConnectionState_Set(clCDeviceDelegateDeviceConnectionState delegate,
                                                 clCDeviceHandlerConnectionState callback) NOEXCEPT;

/**
 * Connect to Neiry device. Non-blocking - OnConnectionStateChanged event is
 * fired when connected.
 *
 * \param device device handle
 */
CL_DLL void clCDevice_Connect(clCDevice device) NOEXCEPT;
/**
 * Disconnect from Neiry device. Non-blocking - OnConnectionStateChanged event
 * is fired when disconnected.
 *
 * \param device device handle
 */
CL_DLL void clCDevice_Disconnect(clCDevice device) NOEXCEPT;
/**
 * \brief Get device battery charge
 *
 * \param device
 * \return battery charge
 */
CL_DLL clCUChar clCDevice_GetBatteryCharge(clCDevice device) NOEXCEPT;
/**
 * \brief Switch device from signal to resistance mode or switch the device off
 *
 * \param device
 * \param mode
 */
CL_DLL void clCDevice_SwitchMode(clCDevice device, clCDeviceMode mode) NOEXCEPT;
CL_DLL clCDeviceMode clCDevice_GetMode(clCDevice device) NOEXCEPT;

/**
 * Is Neiry device connected to Capsule.
 *
 * \param device device handle
 */
CL_DLL bool clCDevice_IsConnected(clCDevice device) NOEXCEPT;
/**
 * Get device information.
 *
 * \param device device handle
 * \returns device info handle
 */
CL_DLL clCDeviceInfo clCDevice_GetInfo(clCDevice device) NOEXCEPT;
/**
 * Get delegate for device resistances event.
 *
 * \param device device handle
 */
CL_DLL clCDeviceDelegateResistances
clCDevice_GetOnResistancesEvent(clCDevice device) NOEXCEPT;
/**
 * Get delegate for device battery charge event.
 *
 * \param device device handle
 */
CL_DLL clCDeviceDelegateUChar clCDevice_GetOnBatteryChargeEvent(clCDevice device) NOEXCEPT;
CL_DLL clCDeviceDelegateMode clCDevice_GetOnModeSwitchedEvent(clCDevice device) NOEXCEPT;
/**
 * Release the device.
 *
 * \param device device handle
 */
CL_DLL void clCDevice_Release(clCDevice device) NOEXCEPT;

typedef enum clCUserActivity {
    clCUserActivity1,
    clCUserActivity2,
    clCUserActivity3,
    clCUserActivity4,
    clCUserActivity5,
    clCUserActivityNone
} clCUserActivity;
/**
 * Destroy Capsule session. Release the handle.
 *
 * \param session session handle
 */
CL_DLL void clCSession_Destroy(clCSession session) NOEXCEPT;
/**
 * Get numeric ID of the session.
 *
 * \param session session handle
 * \returns int64 ID of the session
 */
CL_DLL int64_t clCSession_GetSessionID(clCSession session) NOEXCEPT;
/**
 * Get UUID of the session.
 *
 * \param session session handle
 * \returns string UUID of the session
 */
CL_DLL clCString clCSession_GetSessionUUID(clCSession session) NOEXCEPT;
/**
 * Get session state.
 *
 * \param session session handle
 */
CL_DLL clCSessionState clCSession_GetSessionState(clCSession session) NOEXCEPT;
/**
 * Start the session.
 *
 * \param session session handle
 */
CL_DLL void clCSession_Start(clCSession session) NOEXCEPT;
/**
 * Stop the session.
 *
 * \param session session handle
 */
CL_DLL void clCSession_Stop(clCSession session) NOEXCEPT;
/**
 * Is the session active.
 *
 * \param session session handle
 */
CL_DLL bool clCSession_IsActive(clCSession session) NOEXCEPT;

/**
 * Set extra data to send to platform
 *
 * \param session session handle
 * \param json extra data in json format
 */
CL_DLL void clCSession_SetExtra(clCSession session, const char* json) NOEXCEPT;

/**
 * \brief Mark user activity start/stop
 *
 * \param session
 * \param marker user activity marker or None for the end of user activity
 */
CL_DLL void clCSession_MarkActivity(clCSession session, clCUserActivity marker) NOEXCEPT;
/**
 * \brief Check if EEG in the session is filtered
 *
 * \param session
 * \return \c true if EEG is filtered
 */
CL_DLL bool clCSession_IsFilterEEG(clCSession session) NOEXCEPT;

/**
 * \brief Configure whether channels in the session should work in bipolar mode
 *
 * \param session
 * \param bipolarMode \c true if bipolar mode is enabled
 * \param filterEEG \c true if raw EEG should be filtered
 */
CL_DLL void clCSession_SetConfig(clCSession session, bool bipolarMode, bool filterEEG) NOEXCEPT;
/**
 * \brief Check if bipolar mode is enabled for the session
 *
 * \param session
 * \return \c true if bipolar mode is enabled
 */
CL_DLL bool clCSession_IsBipolarMode(clCSession session) NOEXCEPT;

/**
 * Session delegate without parameters. Not supported in the current version
 */
CLC_STRUCT(clCSessionDelegate);
typedef void (*clCSessionHandler)(clCSession);
CL_DLL void clCSessionDelegate_Set(clCSessionDelegate delegate,
                                   clCSessionHandler callback) NOEXCEPT;

/**
 * Get session delegate for started event.
 *
 * \param session session handle
 */
CL_DLL clCSessionDelegate
clCSession_GetOnSessionStartedEvent(clCSession session) NOEXCEPT;
/**
 * Get session delegate for stopped event.
 *
 * \param session session handle
 */
CL_DLL clCSessionDelegate
clCSession_GetOnSessionStoppedEvent(clCSession session) NOEXCEPT;

/**
 * Get session delegate for error event.
 *
 * \param session session handle
 */
CL_DLL clCSessionDelegateSessionError
clCSession_GetOnErrorEvent(clCSession session) NOEXCEPT;

/**
 * \brief Individual NFB calibrator
 */
CLC_CLASS_WN(NFBCalibratorPrivate, clCNFBCalibrator);

/**
 * \brief NFB call result.
 */
typedef enum clCNFBCallResult {
    clC_NFB_Success,           /**< Call has finished successfully. */
    clC_NFB_NotInitialized,    /**< NFB interface has not been initialized. */
    clC_NFB_ModelIsNotTrained, /**< Model was not trained, must be trained
                                  first. */
    clC_NFB_ModelIsTrained,    /**< Model is already trained, does not need any
                                  training. */
    clC_NFB_FailedToSendData   /**< Failed to send data, session might not be
                                  active. */
} clCNFBCallResult;

/**
 * \brief NFB state.
 */
typedef enum ccCNFBState {
    cc_NFB_Undefined,
    cc_NFB_Relaxation,
    cc_NFB_Concentration
} ccCNFBState;

/**
 * \brief User state, determined by NFB classifier.
 */
typedef struct clCNFBUserState {
    /**
     * \brief User's neural feedback.
     */
    const float* feedbackData;
    uint64_t feedbackSize;
    int64_t timestamp;
} clCNFBUserState;

/**
 * \brief Interface to NFB classification process.
 *
 * Before calling any of the functions make sure that:
 * * A valid \ref clCSession created with clCClient_CreateSession is passed to
 * the constructor.
 * * \ref clCSession is active.
 *
 * Model might need to be trained before it is able to calculate user state.
 * clCNFB_IsModelTrained will return \c false if it needs to be trained.
 *
 * In order to train model you should:
 * * Mark each state at least once, with duration no less than a time window
 */
CLC_CLASS_WN(ClassificationNFBPrivate, clCNFB);
CLC_STRUCT(clCNFBDelegateNFBUserState);
typedef void (*clCNFBHandlerNFBUserState)(clCNFB, const clCNFBUserState*);
CL_DLL void clCNFBDelegateNFBUserState_Set(clCNFBDelegateNFBUserState delegate,
                                           clCNFBHandlerNFBUserState callback) NOEXCEPT;
/**
 * Create a NFB classifier.
 *
 * \param session session handle. Valid session must be passed (does not need to be active)
 * \returns NFB handle.
 */
CL_DLL clCNFB clCNFB_Create(clCSession session) NOEXCEPT;
/**
 * Create a NFB classifier after individual NFB has calibrated.
 *
 * \param calibrator NFB calibrator handle. A calibrated calibrator must be passed.
 * \returns NFB handle.
 */
CL_DLL clCNFB clCNFB_CreateCalibrated(clCNFBCalibrator calibrator, clCError* error) NOEXCEPT;
/**
 * Destroy NFB classifier. Release the handle.
 *
 * \param nfb NFB handle
 */
CL_DLL void clCNFB_Destroy(clCNFB nfb) NOEXCEPT;

CLC_STRUCT(clCNFBDelegate);
typedef void (*clCNFBHandler)(clCNFB);
CL_DLL void clCNFBDelegate_Set(clCNFBDelegate delegate, clCNFBHandler callback) NOEXCEPT;

CLC_STRUCT(clCNFBDelegateBool);
typedef void (*clCNFBHandlerBool)(clCNFB, bool);
CL_DLL void clCNFBDelegateBool_Set(clCNFBDelegateBool delegate,
                                   clCNFBHandlerBool callback) NOEXCEPT;

CLC_STRUCT(clCNFBDelegateString);
typedef void (*clCNFBHandlerString)(clCNFB, const char*);
CL_DLL void clCNFBDelegateString_Set(clCNFBDelegateString delegate,
                                     clCNFBHandlerString callback) NOEXCEPT;

CL_DLL clCNFBDelegate clCNFB_GetOnInitializedEvent(clCNFB nfb) NOEXCEPT;
CL_DLL clCNFBDelegate clCNFB_GetOnModelTrainedEvent(clCNFB nfb) NOEXCEPT;
CL_DLL clCNFBDelegateBool clCNFB_GetOnResetEvent(clCNFB nfb) NOEXCEPT;
CL_DLL clCNFBDelegateNFBUserState clCNFB_GetOnUserStateChangedEvent(clCNFB nfb) NOEXCEPT;
CL_DLL clCNFBDelegateString clCNFB_GetOnErrorEvent(clCNFB nfb) NOEXCEPT;

/**
 * \brief Request initialization.
 *
 * OnInitialized event is called on completion.
 *
 * \param nfb NFB handle
 *
 * \returns call result.
 */
CL_DLL clCNFBCallResult clCNFB_Initialize(clCNFB nfb) NOEXCEPT;
/**
 * \brief Request initialization.
 *
 * OnInitialized event is called on completion.
 *
 * \param nfb NFB handle
 * \param userStateCheckPeriod period of user state updates
 *
 * \returns call result.
 */
CL_DLL clCNFBCallResult clCNFB_InitializeWithPeriod(clCNFB nfb, float userStatePeriod) NOEXCEPT;
/**
 * \brief Add calculation of neuro feedback with configured output.
 *
 * \param nfb NFB handle
 * \param feedbackFunction function of bands, e.g. "(alpha + beta) / theta"
 *
 * \returns call result.
 */
CL_DLL clCNFBCallResult clCNFB_AddFeedbackFunction(clCNFB nfb, const char* feedbackFunction) NOEXCEPT;

/**
 * \brief Request training, event is called when training is finished
 *
 * OnModelTrained event is called on completion
 *
 * \param nfb NFB handle
 *
 * \returns call result.
 */
CL_DLL clCNFBCallResult clCNFB_Train(clCNFB nfb) NOEXCEPT;
/**
 * \brief Request reset, event is called when reset finished
 *
 * OnReset event is called on completion
 *
 * \param nfb NFB handle
 *
 * \returns call result.
 */
CL_DLL clCNFBCallResult clCNFB_Reset(clCNFB nfb) NOEXCEPT;
/**
 * \brief Mark a change in user state
 *
 * Used in BCI system to train classifier to discriminate states.
 * Should always end with Undefined state before Train() to mark an ending.
 *
 * \param nfb NFB handle
 * \param userState state to be marked
 *
 * \returns call result.
 */
// DEPRECATED
CL_DLL clCNFBCallResult clCNFB_MarkState(clCNFB nfb, ccCNFBState userState) NOEXCEPT;

/**
 * \brief Check whether interface was initialized.
 *
 * \param nfb NFB handle
 *
 * \returns whether interface is initialized.
 */
CL_DLL bool clCNFB_IsInitialized(clCNFB nfb) NOEXCEPT;
/**
 * \brief Check whether model was trained.
 *
 * \param nfb NFB handle
 *
 * \returns whether model is trained.
 */
CL_DLL bool clCNFB_IsModelTrained(clCNFB nfb) NOEXCEPT;

CLC_CLASS_WN(NFBBaselinePrivate, clCNFBBaseline);
CL_DLL clCNFBBaseline clCNFBBaseline_Create(float windowSizeSec,
                                            float strideSec) NOEXCEPT;
CL_DLL void clCNFBBaseline_Destroy(clCNFBBaseline baseline) NOEXCEPT;
CL_DLL void clCNFBBaseline_PushNFB(clCNFBBaseline baseline, float nfb,
                                   int64_t timestamp) NOEXCEPT;
CL_DLL float clCNFBBaseline_Get(clCNFBBaseline baseline, uint8_t percentile) NOEXCEPT;

/**
 * \brief Wrapper of \ref clCNFB
 *
 * Use it for metrics described in \ref clCEmotionsEmotion enumeration. **Do
 * not use it at the same time as using \ref clCNFB.**
 */
CLC_CLASS_WN(ClassificationEmotionsPrivate, clCEmotions);

/**
 * \brief Metrics you can compute with \ref clCEmotions
 */
typedef enum clCEmotionsEmotion {
    clC_Emotions_EmotionFocus,
    clC_Emotions_EmotionChill,
    clC_Emotions_EmotionStres,
    clC_Emotions_EmotionAnger,
    clC_Emotions_EmotionSelfControl,
} clCEmotionsEmotion;

/**
 * \brief Emotions values, determined by Emotions classifier.
 */
typedef struct clCEmotionalStates {
    float focus;
    float chill;
    float stress;
    float anger;
    float selfControl;
} clCEmotionalStates;

/**
 * Create an Emotions classifier.
 * Start calibration and update metrics after NFB calibration finished.
 *
 * \param session session handle. Valid session must be passed (does not need to be active)
 * \return Emotions handle.
 */
CL_DLL clCEmotions clCEmotions_Create(clCSession session, double speed, double maxSpeed) NOEXCEPT;
/**
 * Create an Emotions classifier after individual NFB has calibrated.
 *
 * \param calibrator NFB calibrator handle. A calibrated calibrator must be passed.
 * \return Emotions handle.
 */
CL_DLL clCEmotions clCEmotions_CreateCalibrated(clCNFBCalibrator calibrator, double speed, double maxSpeed) NOEXCEPT;
/**
 * Destroy Emotions classifier.
 *
 * \param emotions Emotions handle.
 */
CL_DLL void clCEmotions_Destroy(clCEmotions emotions) NOEXCEPT;

CLC_STRUCT(clCEmotionsDelegateFloat);
typedef void (*clCEmotionsHandlerFloat)(clCEmotions, float);
CL_DLL void clCEmotionsDelegateFloat_Set(clCEmotionsDelegateFloat delegate,
                                         clCEmotionsHandlerFloat handler) NOEXCEPT;
CL_DLL clCEmotionsDelegateFloat clCEmotions_GetOnEmotionUpdateEvent(
    clCEmotions emotions, clCEmotionsEmotion emotion) NOEXCEPT;

CLC_STRUCT(clCEmotionsDelegateEmotionalStatesUpdate);
typedef void (*clCEmotionsHandlerEmotionalStatesUpdate)(clCEmotions, const clCEmotionalStates*);
CL_DLL void clCEmotionsDelegateEmotionalStatesUpdate_Set(clCEmotionsDelegateEmotionalStatesUpdate delegate,
                                                         clCEmotionsHandlerEmotionalStatesUpdate handler) NOEXCEPT;
CL_DLL clCEmotionsDelegateEmotionalStatesUpdate clCEmotions_GetOnEmotionalStatesUpdateEvent(clCEmotions emotions) NOEXCEPT;

CLC_STRUCT(clCEmotionsDelegate);
typedef void (*clCEmotionsHandler)(clCEmotions);
CL_DLL void clCEmotionsDelegate_Set(clCEmotionsDelegate delegate, clCEmotionsHandler handler) NOEXCEPT;
CL_DLL clCEmotionsDelegate clCEmotions_GetOnCalibratedEvent(clCEmotions emotions) NOEXCEPT;

/**
 * \brief Initialize emotions classifier
 *
 * \param emotions
 * \param platformAddress address of platform to upload to
 */
CL_DLL void clCEmotions_Initialize(clCEmotions emotions, const char* platformAddress) NOEXCEPT;
/**
 * \brief Start emotions calibration
 *
 * \param emotions
 */
CL_DLL void clCEmotions_StartCalibration(clCEmotions emotions) NOEXCEPT;

/**
 * \brief Create a Cardio classifier.
 *
 * \param session session handle. Valid session must be passed (does not need to be active)
 * \return Cardio classifier handle.
 */
CL_DLL clCCardio clCCardio_Create(clCSession session) NOEXCEPT;
/**
 * \brief Create a Cardio classifier after individual NFB has calibrated.
 *
 * \param calibrator NFB calibrator handle. A calibrated calibrator must be passed.
 * \return Cardio handle.
 */
CL_DLL clCCardio clCCardio_CreateCalibrated(clCNFBCalibrator calibrator) NOEXCEPT;

/**
 * \brief Initialize cardio classifier
 *
 * \param cardio classifier handle
 */
CL_DLL void clCCardio_Initialize(clCCardio cardio) NOEXCEPT;

CLC_STRUCT(clCCardioIndexesDelegate);
typedef struct clCCardioData {
    float heartRate;
    float stressIndex;
    float caplanIndex;
    bool artifacted;
} clCCardioData;
typedef void (*clCCardioIndexesUpdateHandler)(clCCardio, clCCardioData);
/**
 * \brief Get delegate for cardio indexes event
 */
CL_DLL clCCardioIndexesDelegate clCCardio_GetOnIndexesUpdateEvent(clCCardio cardio) NOEXCEPT;
/**
 * \brief Set handler for cardio indexes updates delegate
 */
CL_DLL void clCCardioDelegateIndexesUpdate_Set(clCCardioIndexesDelegate delegate,
                                               clCCardioIndexesUpdateHandler callback) NOEXCEPT;
/**
 * \brief Destroy Cardio classifier.
 *
 * \param cardio classifier handle.
 */
CL_DLL void clCCardio_Destroy(clCCardio cardio) NOEXCEPT;
/**
 * \brief Create a MEMS classifier.
 *
 * \param session session handle. Valid session must be passed (does not need to be active)
 * \return MEMS classifier handle.
 */
CL_DLL clCMEMS clCMEMS_Create(clCSession session) NOEXCEPT;
/**
 * \brief Create a MEMS classifier after individual NFB has calibrated.
 *
 * \param calibrator NFB calibrator handle. A calibrated calibrator must be passed.
 * \return MEMS handle.
 */
CL_DLL clCMEMS clCMEMS_CreateCalibrated(clCNFBCalibrator calibrator) NOEXCEPT;

/**
 * \brief Initialize cardio classifier
 *
 * \param mems classifier handle
 */
CL_DLL void clCMEMS_Initialize(clCMEMS mems) NOEXCEPT;

CLC_STRUCT(clCMEMSTimedDataDelegate);
/**
 * \brief Get delegate for MEMS data event
 */
CL_DLL clCMEMSTimedDataDelegate clCMEMS_GetOnMEMSTimedDataUpdateEvent(clCMEMS mems) NOEXCEPT;
/**
 * @brief Array of MEMS data with timepoints
 */
CLC_STRUCT_WN(MEMSTimedData, clCMEMSTimedData);
/**
 * @brief Three-dimensional point
 */
typedef struct clCPoint3d {
    float x;
    float y;
    float z;
} clCPoint3d;
/**
 * Get total number of MEMS values.
 *
 * \param memsTimedData mems values
 * \returns number of values
 */
CL_DLL int32_t clCMEMSTimedData_GetCount(clCMEMSTimedData memsTimedData) NOEXCEPT;
/**
 * Get accelerometer orientation by index.
 *
 * \param memsTimedData mems values
 * \param index index
 * \returns accelerometer orientation
 */
CL_DLL clCPoint3d clCMEMSTimedData_GetAccelerometer(clCMEMSTimedData memsTimedData,
                                                    int32_t index) NOEXCEPT;
/**
 * Get gyroscope orientation by index.
 *
 * \param memsTimedData mems values
 * \param index index
 * \returns gyroscope orientation
 */
CL_DLL clCPoint3d clCMEMSTimedData_GetGyroscope(clCMEMSTimedData memsTimedData,
                                                int32_t index) NOEXCEPT;
/**
 * Get MEMS time point by index.
 *
 * \param memsTimedData mems values
 * \param index index
 * \returns time point
 */
CL_DLL float clCMEMSTimedData_GetTimepoint(clCMEMSTimedData memsTimedData, int32_t index) NOEXCEPT;

typedef void (*clCMEMSDataUpdateHandler)(clCMEMS, clCMEMSTimedData);
/**
 * \brief Set handler for MEMS data updates delegate
 */
CL_DLL void clCMEMSDelegateMEMSTimedDataUpdate_Set(clCMEMSTimedDataDelegate delegate,
                                                   clCMEMSDataUpdateHandler callback) NOEXCEPT;

/**
 * \brief Destroy MEMS classifier.
 *
 * \param mems classifier handle.
 */
CL_DLL void clCMEMS_Destroy(clCMEMS mems) NOEXCEPT;

CLC_CLASS_WN(ProductivityMetricsPrivate, clCNFBMetricProductivity);
/**
 * \brief Create a NFB classifier for metrics productivity
 *
 * \param session session handle. Valid session must be passed (does not need to
 * be active)
 * \param path path to the directory with logs
 * \param error out error parameter
 *
 * \returns NFBMetrics productivity handle
 */
CL_DLL clCNFBMetricProductivity clCNFBMetricsProductivity_Create(
    clCSession session, const char* path, double speed, double maxSpeed,
    double slowDown, clCError* error) NOEXCEPT;
/**
 * \brief Create a NFB classifier for metrics productivity after individual NFB has been calibrated
 *
 * \param calibrator calibrator handle. A calibrated calibrator must be passed
 * \param path path to the directory with logs
 * \param error out error parameter
 *
 * \returns NFBMetrics productivity handle
 */
CL_DLL clCNFBMetricProductivity clCNFBMetricsProductivity_CreateCalibrated(
    clCNFBCalibrator calibrator, const char* path, double speed, double maxSpeed,
    double slowDown, clCError* error) NOEXCEPT;
/**
 * Destroy NFB classifier. Release the NFBMetrics productivity handle.
 *
 * \param nfbMetrics NFBMetrics productivity handle
 */
CL_DLL void
clCNFBMetricsProductivity_Destroy(clCNFBMetricProductivity nfbMetrics) NOEXCEPT;
/**
 * \brief Initialize NFB classifier for metrics productivity
 *
 * \param nfbMetrics NFBMetrics productivity handle
 */
CL_DLL void
clCNFBMetricsProductivity_InitializeNFB(clCNFBMetricProductivity nfbMetrics, const char* platformAddress,
                                        clCError* error) NOEXCEPT;

CLC_STRUCT(clCNFBMetricsProductivityCalibratedDelegate);
typedef void (*clCNFBMetricsProductivityCalibratedHandler)(
    clCNFBMetricProductivity);
/**
 * \brief Set handler for calibrated event NFB classifier, not involved yet
 */
CL_DLL void clCNFBMetricsProductivity_CalibratedEvent_Set(
    clCNFBMetricsProductivityCalibratedDelegate delegate,
    clCNFBMetricsProductivityCalibratedHandler handler) NOEXCEPT;

/**
 * \brief Get delegate for baseline calibration finished event
 */
CL_DLL clCNFBMetricsProductivityCalibratedDelegate
clCNFBMetricsProductivity_GetOnBaselineCalibratedEvent(clCNFBMetricProductivity nfbMetrics) NOEXCEPT;

typedef struct clCNFBMetricsProductivityBaselines {
    float gravityBaseline;
    float productivityBaseline;
    float fatigueBaseline;
    float reverseFatigueBaseline;
    float relaxBaseline;
    float concentrationBaseline;
} clCNFBMetricsProductivityBaselines;
CLC_STRUCT(clCNFBMetricsProductivityInitialBaselineCalibratedDelegate);
typedef void (*clCNFBMetricsProductivityInitialBaselineCalibratedHandler)(
    clCNFBMetricProductivity, const clCNFBMetricsProductivityBaselines*);
/**
 * \brief Set delegate for initial baseline calibration finished event
 */
CL_DLL void clCNFBMetricsProductivity_InitialBaselineCalibratedEvent_Set(
    clCNFBMetricsProductivityInitialBaselineCalibratedDelegate delegate,
    clCNFBMetricsProductivityInitialBaselineCalibratedHandler handler) NOEXCEPT;

/**
 * \brief Get delegate for initial baseline calibration finished event
 */
CL_DLL clCNFBMetricsProductivityInitialBaselineCalibratedDelegate
clCNFBMetricsProductivity_GetOnInitialBaselineCalibratedEvent(clCNFBMetricProductivity nfbMetrics) NOEXCEPT;

typedef struct clCNFBMetricsProductivityIndividualIndexes {
    float IAF;
    float IAPF;
    float IAPFPowerNormalized;
    float alphaRangeBegin;
    float alphaRangeEnd;
    uint8_t relaxPoint;
    uint8_t stressPoint;
    float relaxBaseline;
    float fatigueBaseline;
    float gravityBaseline;
    float concentrationBaseline;
    int isArtifacted;
} clCNFBMetricsProductivityIndividualIndexes;

CLC_STRUCT(clCNFBMetricsProductivityIndividualDelegate);
typedef void (*clCNFBMetricsProductivityIndividualHandler)(clCNFBMetricProductivity, const clCNFBMetricsProductivityIndividualIndexes*);

/**
 * \brief Get delegate for Individual measure event
 */
CL_DLL clCNFBMetricsProductivityIndividualDelegate
clCNFBMetricsProductivity_GetOnIndividualMeasuredEvent(clCNFBMetricProductivity nfbMetrics) NOEXCEPT;
/**
 * \brief Set handler for Individual measure event
 */
CL_DLL void clCNFBMetricsProductivity_IndividualMeasuredEvent_Set(
    clCNFBMetricsProductivityIndividualDelegate delegate,
    clCNFBMetricsProductivityIndividualHandler handler) NOEXCEPT;

typedef struct clCNFBMetricsProductivityValues {
    float fatigueScore;
    float gravityScore;
    float concentrationScore;
} clCNFBMetricsProductivityValues;

CLC_STRUCT(clCNFBMetricsProductivityValuesDelegate);
typedef void (*clCNFBMetricsProductivityValuesHandler)(clCNFBMetricProductivity, const clCNFBMetricsProductivityValues*);
/**
 * \brief Get delegate for Productivity scores event
 */
CL_DLL clCNFBMetricsProductivityValuesDelegate
clCNFBMetricsProductivity_GetOnProductivityValuesEvent(clCNFBMetricProductivity nfbMetrics) NOEXCEPT;
/**
 * \brief Set handler for Productivity scores event
 */
CL_DLL void clCNFBMetricsProductivity_ValuesEvent_Set(
    clCNFBMetricsProductivityValuesDelegate delegate,
    clCNFBMetricsProductivityValuesHandler handler) NOEXCEPT;

/**
 * \brief Set handler for Individual measure event
 *
 * \returns IAPF or 0 if until the measurement individual state is completed
 */
CL_DLL float clCNFBMetricsProductivity_GetIAPF(
    clCNFBMetricProductivity nfbMetrics) NOEXCEPT;
/**
 * \brief Set handler for Individual measure event
 *
 * \returns IAF or 0 if until the measurement individual state is completed
 */
CL_DLL float
clCNFBMetricsProductivity_GetIAF(clCNFBMetricProductivity nfbMetrics) NOEXCEPT;
/**
 * \brief Artifacts when measuring the state user
 */
typedef struct clCNFBUserArtifacts {
    /**
     * \brief Artifacts from channel.
     */
    const uint8_t* artifactsData;
    uint64_t artifactsSize;
    int64_t timestamp;
} clCNFBUserArtifacts;

CLC_STRUCT(clCNFBMetricsProductivityArtifactsDelegate);
typedef void (*clCNFBMetricsProductivityArtifactsHandler)(
    clCNFBMetricProductivity, const clCNFBUserArtifacts*);
/**
 * \brief Get delegate for occurrence artifacts event
 */
CL_DLL clCNFBMetricsProductivityArtifactsDelegate
clCNFBMetricsProductivity_GetOnArtifactsEvent(
    clCNFBMetricProductivity nfbMetrics) NOEXCEPT;
/**
 * \brief Set handler for occurrence artifacts event
 */
CL_DLL void clCNFBMetricsProductivity_ArtifactsEvent_Set(
    clCNFBMetricsProductivityArtifactsDelegate delegate,
    clCNFBMetricsProductivityArtifactsHandler handler) NOEXCEPT;

//------------------
typedef float clCNFBMetricsProductivity_CurrentPowerIAPF,
    clCNFBMetricsProductivity_relaxValue;

CLC_STRUCT(clCNFBMetricsProductivityTrainingInfoDelegate);
typedef void (*clCNFBMetricsProductivityTrainingInfoHandler)(
    clCNFBMetricProductivity, clCNFBMetricsProductivity_CurrentPowerIAPF,
    clCNFBMetricsProductivity_relaxValue);

CL_DLL void clCNFBMetricsProductivity_TrainingInfoEvent_Set(
    clCNFBMetricsProductivityTrainingInfoDelegate delegate,
    clCNFBMetricsProductivityTrainingInfoHandler handler) NOEXCEPT;

CL_DLL clCNFBMetricsProductivityTrainingInfoDelegate
clCNFBMetricsProductivity_GetOnTrainingInfoEvent(clCNFBMetricProductivity nfbMetrics) NOEXCEPT;

//----------------

CLC_STRUCT(clCNFBMetricsProductivityEventDelegate);
typedef void (*clCNFBMetricsProductivityEventHandler)(clCNFBMetricProductivity, float);
/**
 * \brief Set handler for update event productivity metrics: 1min, 5min ...
 * 60min
 */
CL_DLL void clCNFBMetricsProductivity_UpdateEvent_Set(
    clCNFBMetricsProductivityEventDelegate delegate,
    clCNFBMetricsProductivityEventHandler handler) NOEXCEPT;
/**
 * \brief Get delegate for update event productivity metric 1 min.
 */
CL_DLL clCNFBMetricsProductivityEventDelegate
clCNFBMetricsProductivity_GetOnUpdateEvent_1min(
    clCNFBMetricProductivity nfbMetrics) NOEXCEPT;
/**
 * \brief Get delegate for update event productivity metric 5 min.
 */
CL_DLL clCNFBMetricsProductivityEventDelegate
clCNFBMetricsProductivity_GetOnUpdateEvent_5min(
    clCNFBMetricProductivity nfbMetrics) NOEXCEPT;
/**
 * \brief Get delegate for update event productivity metric 15 min.
 */
CL_DLL clCNFBMetricsProductivityEventDelegate
clCNFBMetricsProductivity_GetOnUpdateEvent_15min(
    clCNFBMetricProductivity nfbMetrics) NOEXCEPT;
/**
 * \brief Get delegate for update event productivity metric 30 min.
 */
CL_DLL clCNFBMetricsProductivityEventDelegate
clCNFBMetricsProductivity_GetOnUpdateEvent_30min(
    clCNFBMetricProductivity nfbMetrics) NOEXCEPT;
/**
 * \brief Get delegate for update event productivity metric 60 min.
 */
CL_DLL clCNFBMetricsProductivityEventDelegate
clCNFBMetricsProductivity_GetOnUpdateEvent_60min(
    clCNFBMetricProductivity nfbMetrics) NOEXCEPT;

/**
 * \brief Obtain an individual NFB calibrator for the session. If not present, it'll
 * be created.
 *
 * \param session session handle. A valid session must be passed (does not need to
 * be active)
 *
 * \returns NFB calibrator handle
 */
CL_DLL clCNFBCalibrator clCNFBCalibrator_CreateOrGet(clCSession session) NOEXCEPT;

enum clCIndividualNFBCalibrationStage {
    clCIndividualNFBCalibrationStage1,
    clCIndividualNFBCalibrationStage2,
    clCIndividualNFBCalibrationStage3,
    clCIndividualNFBCalibrationStage4,
};
/**
 * \brief Start stage of individual NFB calibration
 *
 * \param stage the user must call this method 4 times sequentially
 * changing this argument from Stage1 to Stage4
 * \param error the error out parameter
 */
CL_DLL void clCNFBCalibrator_CalibrateIndividualNFB(clCNFBCalibrator nfbCalibrator,
                                                    clCIndividualNFBCalibrationStage stage, clCError* error) NOEXCEPT;

/**
 * \brief Start stage of quick individual NFB calibration
 *
 * \param error the error out parameter
 */
CL_DLL void clCNFBCalibrator_CalibrateIndividualNFBQuick(clCNFBCalibrator nfbCalibratorPtr,
                                                         clCError* error) NOEXCEPT;

CLC_STRUCT(clCNFBCalibratorDelegateIndividualNFBCalibrationStage);
typedef void (*clCNFBCalibratorHandlerIndividualNFBCalibrationStage)(clCNFBCalibrator);
/**
 * \brief Set handler for individual calibration event
 */
CL_DLL void clCNFBCalibratorDelegateIndividualNFBCalibrationStage_Set(
    clCNFBCalibratorDelegateIndividualNFBCalibrationStage delegate,
    clCNFBCalibratorHandlerIndividualNFBCalibrationStage handler) NOEXCEPT;
/**
 * \brief Get handler for individual calibration stage event
 *
 * \param nfbCalibrator
 * \return calibration stage delegate
 */
CL_DLL clCNFBCalibratorDelegateIndividualNFBCalibrationStage
clCNFBCalibrator_GetOnIndividualNFBStageFinishedEvent(clCNFBCalibrator nfbCalibrator) NOEXCEPT;

struct clCIndividualNFBData {
    /**
     * \brief Individual NFB frequency.
     */
    float individualFrequency = 0.0f;
    /**
     * \brief Individual NFB peak frequency.
     */
    float individualPeakFrequency = 0.0f;
    /**
     * \brief Individual NFB peak frequency power.
     */
    float individualPeakFrequencyPower = 0.0f;
    /**
     * \brief Individual NFB peak frequency suppression.
     */
    float individualPeakFrequencySuppression = 0.0f;
    /**
     * \brief Individual NFB bandwidth.
     */
    float individualBandwidth = 0.0f;
    /**
     * \brief NFB power.
     */
    float individualNormalizedPower = 0.0f;
    /**
     * \brief Left frequency bound
     */
    float lowerFrequency = 0.0f;
    /**
     * \brief Right frequency bound
     */
    float upperFrequency = 0.0f;
};

CLC_STRUCT(clCNFBCalibratorDelegateIndividualNFBCalibrated);
typedef void (*clCNFBCalibratorHandlerIndividualNFBCalibrated)(
    clCNFBCalibrator, const clCIndividualNFBData*);
/**
 * \brief Set handler for individual calibration finish event
 */
CL_DLL void clCNFBCalibratorDelegateIndividualNFBCalibrated_Set(
    clCNFBCalibratorDelegateIndividualNFBCalibrated delegate,
    clCNFBCalibratorHandlerIndividualNFBCalibrated handler) NOEXCEPT;
/**
 * \brief Get handler for individual calibration finish event
 *
 * \param nfbCalibrator
 * \return calibration finish delegate
 */
CL_DLL clCNFBCalibratorDelegateIndividualNFBCalibrated
clCNFBCalibrator_GetOnIndividualNFBCalibratedEvent(clCNFBCalibrator nfbCalibrator) NOEXCEPT;

/**
 * \brief Is calibration finished
 *
 * \param nfbCalibrator
 * \return \c true if calibration has finished
 */
CL_DLL bool clCNFBCalibrator_IsCalibrated(clCNFBCalibrator nfbCalibrator) NOEXCEPT;
/**
 * \brief Has calibration failed
 *
 * \param nfbCalibrator
 * \return \c true if calibration has failed
 */
CL_DLL bool clCNFBCalibrator_HasCalibrationFailed(clCNFBCalibrator nfbCalibrator) NOEXCEPT;

#ifdef __cplusplus
}
#endif
